# -*- coding: utf-8 -*-
"""
Copyright (C) 2012-2022  Diego Torres Milano
Created on Feb 2, 2015

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Diego Torres Milano
"""

from __future__ import print_function

__version__ = '22.0.0'

import os
import platform
import re
import subprocess
import sys
import threading
import time
import warnings
from abc import ABC
from datetime import datetime
from typing import Optional, List, Tuple

import culebratester_client
from culebratester_client import Text, ObjectRef, DefaultApi, Point, PerformTwoPointerGestureBody, \
    BooleanResponse, NumberResponse, StatusResponse

from com.dtmilano.android.adb.adbclient import AdbClient
from com.dtmilano.android.common import obtainAdbPath

__author__ = 'diego'

DEBUG = False

lock = threading.Lock()


class RunTestsThread(threading.Thread):
    """
    Runs the instrumentation for the specified package in a new thread.
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, verbose=None, adbClient=None,
                 testClass=None, testRunner=None):
        threading.Thread.__init__(self, group=group, target=target, name=name)
        self.adbClient = adbClient
        self.testClass = testClass
        self.testRunner = testRunner
        self.pkg = re.sub('\\.test$', '', self.testClass)

    def run(self):
        if DEBUG:
            print("RunTestsThread: Acquiring lock", file=sys.stderr)
        lock.acquire()
        if DEBUG:
            print("RunTestsThread: Lock acquired", file=sys.stderr)
        self.forceStop()
        time.sleep(3)
        if DEBUG:
            print("Starting test...", file=sys.stderr)
            print("RunTestsThread: Releasing lock", file=sys.stderr)
        lock.release()
        out = self.adbClient.shell('am instrument -w ' + self.testClass + '/' + self.testRunner + '; echo "ERROR: $?"')
        if DEBUG:
            print("\nFinished test.", file=sys.stderr)
        errmsg = out.splitlines()[-1]
        m = re.match('ERROR: (\\d+)', errmsg)
        if m:
            exitval = int(m.group(1))
            if exitval != 0:
                raise RuntimeError('Cannot start test on device: ' + out)
        else:
            raise RuntimeError('Unknown message')

    def forceStop(self):
        if DEBUG:
            print("Cleaning up before start. Stopping '%s'" % self.pkg, file=sys.stderr)
        self.adbClient.shell('am force-stop ' + self.pkg)


def check_response(response: StatusResponse) -> None:
    """
    Checks if the status response is OK. Otherwise, it raises an exception.
    :param response: the response
    :return: None
    """
    if response.status != 'OK':
        raise RuntimeError(response)


class UiAutomatorHelper:
    PACKAGE = 'com.dtmilano.android.culebratester2'
    TEST_CLASS = PACKAGE + '.test'
    TEST_RUNNER = 'com.dtmilano.android.culebratester2.UiAutomatorHelper'

    def __init__(self, adbclient, adb=None, localport=9987, remoteport=9987, hostname='localhost', api_version='v2'):
        """
        UiAutomatorHelper constructor used when the backend selected is **CulebraTester2-public**.
        This class holds references to the different API's:
         - Device
         - TargetContext
         - ObjectStore
         - UiDevice
         - UiObject
         - UiObject2
         - Until

        :param adbclient: the adb client
        :param adb: adb if known
        :param localport: the local port used in CulebraTester2-public port forwarding
        :param remoteport: the remote port used in CulebraTester2-public port forwarding
        :param hostname: the hostname used by CulebraTester2-public
        :param api_version: the api version
        """
        self.adbClient = adbclient
        ''' The adb client (a.k.a. device) '''
        self.adb = self.__whichAdb(adb)
        ''' The adb command '''
        self.osName = platform.system()
        ''' The OS name. We sometimes need specific behavior. '''
        self.isDarwin = (self.osName == 'Darwin')
        ''' Is it macOS? '''
        self.hostname = hostname
        ''' The hostname we are connecting to. '''

        print(
            f'⚠️  CulebraTester2 server should have been started and localport {localport} redirected to remote port {remoteport}.',
            file=sys.stderr)
        configuration = culebratester_client.Configuration()
        configuration.host = f'http://{hostname}:{localport}/{api_version}'
        self.api_instance: DefaultApi = culebratester_client.DefaultApi(culebratester_client.ApiClient(configuration))
        self.device: UiAutomatorHelper.Device = UiAutomatorHelper.Device(self)
        self.target_context: UiAutomatorHelper.TargetContext = UiAutomatorHelper.TargetContext(self)
        self.object_store: UiAutomatorHelper.ObjectStore = UiAutomatorHelper.ObjectStore(self)
        self.ui_device: UiAutomatorHelper.UiDevice = UiAutomatorHelper.UiDevice(self)
        self.ui_object: UiAutomatorHelper.UiObject = UiAutomatorHelper.UiObject(self)
        self.ui_object2: UiAutomatorHelper.UiObject2 = UiAutomatorHelper.UiObject2(self)
        self.until: UiAutomatorHelper.Until = UiAutomatorHelper.Until(self)

    def __connectSession(self):
        if DEBUG:
            print("UiAutomatorHelper: Acquiring lock", file=sys.stderr)
        lock.acquire()
        if DEBUG:
            print("UiAutomatorHelper: Lock acquired", file=sys.stderr)
            print("UiAutomatorHelper: Connecting session", file=sys.stderr)
        session = None  # requests.Session()
        if not session:
            raise RuntimeError("Cannot create session")
        tries = 10
        while tries > 0:
            time.sleep(0.5)
            if DEBUG:
                print("UiAutomatorHelper: Attempting to connect to", self.baseUrl, '(tries=%s)' % tries,
                      file=sys.stderr)
            try:
                response = session.head(self.baseUrl)
                if response.status_code == 200:
                    break
            except:  # requests.exceptions.ConnectionError as ex:
                tries -= 1
        lock.release()
        if tries == 0:
            raise RuntimeError("Cannot connect to " + self.baseUrl)
        if DEBUG:
            print("UiAutomatorHelper: HEAD", response, file=sys.stderr)
            print("UiAutomatorHelper: Releasing lock", file=sys.stderr)
        # lock.release()
        return session

    @staticmethod
    def __whichAdb(adb):
        if adb:
            if not os.access(adb, os.X_OK):
                raise Exception('adb="%s" is not executable' % adb)
        else:
            # Using adbclient we don't need adb executable yet (maybe it's needed if we want to
            # start adb if not running) or to redirect ports
            adb = obtainAdbPath()

        return adb

    def __redirectPort(self, localport, remoteport):
        self.localPort = localport
        self.remotePort = remoteport
        subprocess.check_call([self.adb, '-s', self.adbClient.serialno, 'forward', 'tcp:%d' % self.localPort,
                               'tcp:%d' % self.remotePort])

    def __runTests(self):
        if DEBUG:
            print("__runTests: start", file=sys.stderr)
        # We need a new AdbClient instance with timeout=None (means, no timeout) for the long running test service
        newAdbClient = AdbClient(self.adbClient.serialno, self.adbClient.hostname, self.adbClient.port, timeout=None)
        self.thread = RunTestsThread(adbClient=newAdbClient, testClass=self.TEST_CLASS, testRunner=self.TEST_RUNNER)
        if DEBUG:
            print("__runTests: starting thread", file=sys.stderr)
        self.thread.start()
        if DEBUG:
            print("__runTests: end", file=sys.stderr)

    def __httpCommand(self, url, params=None, method='GET'):
        raise RuntimeError(f"this method should not be used: url={url} params={params} method={method}")

    @staticmethod
    def timestamp() -> str:
        """
        Timestamp in ISO format.

        WARNING: may not be suitable to include in filenames on some platforms as it may include invalid path chars
        (i.e. `:`).

        :return: the timestamps
        """
        return datetime.now().isoformat()

    #
    # UiAutomatorHelper internal commands
    #
    def quit(self):
        pass

    #
    # API Base
    #
    class ApiBase(ABC):

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__()
            self.uiAutomatorHelper = uiAutomatorHelper

        @staticmethod
        def intersection(l1, l2):
            return list(set(l1) & set(l2))

        def some(self, l1, l2):
            return len(self.intersection(l1, l2)) > 0

        def all(self, l1, l2):
            li = len(self.intersection(l1, l2))
            return li == len(l1) and li == len(l2)

    #
    # Device
    #
    class Device(ApiBase):
        """
        Device
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def display_real_size(self):
            """
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the display real size
            """
            return self.uiAutomatorHelper.api_instance.device_display_real_size_get()

        def dumpsys(self, service, **kwargs) -> str:
            """
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param service: the service
            :return: the dumpsys output
            """
            return self.uiAutomatorHelper.api_instance.device_dumpsys_get(service=service, _preload_content=False,
                                                                          **kwargs).read().decode('UTF-8')

        def get_top_activity_name_and_pid(self) -> Optional[str]:
            dat = self.dumpsys('activity', arg1='top')
            activityRE = re.compile(r'\s*ACTIVITY ([A-Za-z0-9_.]+)/([A-Za-z0-9_.\$]+) \w+ pid=(\d+)')
            m = activityRE.findall(dat)
            if len(m) > 0:
                return m[-1]
            else:
                warnings.warn("NO MATCH:" + dat)
                return None

        def get_top_activity_name(self) -> Optional[str]:
            tanp = self.get_top_activity_name_and_pid()
            if tanp:
                return tanp[0] + '/' + tanp[1]
            else:
                return None

        def get_top_activity_uri(self) -> Optional[str]:
            tan = self.get_top_activity_name()
            dat = self.dumpsys('activity')
            startActivityRE = re.compile(r'^\s*mStartActivity:')
            intentRE = re.compile(f'^\\s*Intent {{ act=(\\S+) dat=(\\S+) flg=(\\S+) cmp={tan} }}')
            lines = dat.splitlines()
            for n, _line in enumerate(lines):
                if startActivityRE.match(_line):
                    for i in range(n, n + 6):
                        m = intentRE.match(lines[i])
                        if m:
                            return m.group(2)
            return None

        def wait_for_new_toast(self, timeout=10000):
            """
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the text in the Toast if found
            """
            return self.uiAutomatorHelper.api_instance.device_wait_for_new_toast_get(timeout=timeout)

    #
    # TargetContext
    #
    class TargetContext(ApiBase):
        """
        Target Context.
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def start_activity(self, pkg, cls, **kwargs):
            """
            Starts an activity.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param pkg: the package
            :param cls: the Activity class
            :param kwargs: uri: the optional URI
            :return: the api response
            """
            return self.uiAutomatorHelper.api_instance.target_context_start_activity_get(pkg, cls, **kwargs)

    #
    # ObjectStore
    #
    class ObjectStore(ApiBase):
        """
        Object Store.
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def clear(self):
            """
            Clears all the objects.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the status
            """
            return self.uiAutomatorHelper.api_instance.object_store_clear_get()

        def list(self):
            """
            List the objects.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the list of objects
            """
            return self.uiAutomatorHelper.api_instance.object_store_list_get()

        def remove(self, oid: int):
            """
            Removes an object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the status
            """
            return self.uiAutomatorHelper.api_instance.object_store_remove_get(oid)

    #
    # UiDevice
    #
    class UiDevice(ApiBase):
        """
        UI Device.
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def click(self, x: int, y: int):
            """
            Clicks on the specified coordinates.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_click_get(x=x, y=y))

        def dump_window_hierarchy(self, _format='JSON'):
            """
            Dumps the window hierarchy.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param _format:
            :return:
            """
            return self.uiAutomatorHelper.api_instance.ui_device_dump_window_hierarchy_get(format=_format)

        def find_object(self, **kwargs):
            """
            Finds an object.
            Invokes GET or POST method depending on the arguments passed.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs:
            :return:
            """
            if 'body' in kwargs:
                return self.uiAutomatorHelper.api_instance.ui_device_find_object_post(**kwargs)
            if self.some(['resource_id', 'ui_selector', 'by_selector'], kwargs):
                return self.uiAutomatorHelper.api_instance.ui_device_find_object_get(**kwargs)
            body = culebratester_client.Selector(**kwargs)
            return self.uiAutomatorHelper.api_instance.ui_device_find_object_post(body=body)

        def find_objects(self, **kwargs):
            """
            Finds objects.
            Invokes GET or POST method depending on the arguments passed.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs:
            :return:
            """
            if 'body' in kwargs:
                return self.uiAutomatorHelper.api_instance.ui_device_find_objects_post(**kwargs)
            if 'by_selector' in kwargs:
                return self.uiAutomatorHelper.api_instance.ui_device_find_objects_get(**kwargs)
            body = culebratester_client.Selector(**kwargs)
            return self.uiAutomatorHelper.api_instance.ui_device_find_objects_post(body=body)

        def has_object(self, **kwargs) -> bool:
            """
            Has an object.
            Invokes GET or POST method depending on the arguments passed.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs:
            :return: True or False
            """
            if 'body' in kwargs:
                return self.uiAutomatorHelper.api_instance.ui_device_has_object_post(**kwargs).value
            if self.some(['resource_id', 'ui_selector', 'by_selector'], kwargs):
                return self.uiAutomatorHelper.api_instance.ui_device_has_object_get(**kwargs).value
            body = culebratester_client.Selector(**kwargs)
            return self.uiAutomatorHelper.api_instance.ui_device_has_object_post(body=body).value

        def press_back(self) -> None:
            """
            Presses BACK.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_press_back_get())

        def press_enter(self) -> None:
            """
            Presses ENTER.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_press_enter_get())

        def press_home(self) -> None:
            """
            Presses HOME.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_press_home_get())

        def press_recent_apps(self) -> None:
            """
            Press recent apps.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_press_recent_apps_get())

        def press_key_code(self, key_code: int, meta_state: int = 0) -> None:
            """
            Presses a key code.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param key_code:
            :param meta_state:
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_press_key_code_get(key_code=key_code,
                                                                                            meta_state=meta_state))

        def swipe(self, **kwargs) -> None:
            """
            Swipes.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs:
            :return:
            """
            if self.all(['start_x', 'start_y', 'end_x', 'end_y', 'steps'], kwargs):
                check_response(self.uiAutomatorHelper.api_instance.ui_device_swipe_get(**kwargs))
                return
            if 'segments' in kwargs:
                # noinspection PyBroadException
                try:
                    segments = kwargs['segments']
                    if isinstance(segments, list):
                        if all(isinstance(e, tuple) for e in segments):
                            kwargs['segments'] = list(map(lambda e: Point(*e), segments))
                except Exception:
                    pass
            body = culebratester_client.SwipeBody(**kwargs)
            check_response(self.uiAutomatorHelper.api_instance.ui_device_swipe_post(body=body))

        def take_screenshot(self, scale=1.0, quality=90, filename=None, **kwargs):
            """
            Takes screenshot.
            Eventually save it in a file specified by filename.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param scale: the scale
            :param quality: the quality
            :param filename: if specified the image will be saved in filename
            :param kwargs: optional arguments
            :return: the response containing the image binary if the filename was not specified
            """
            if filename:
                img = self.uiAutomatorHelper.api_instance.ui_device_screenshot_get(scale=scale,
                                                                                   quality=quality,
                                                                                   _preload_content=False,
                                                                                   **kwargs).read()
                with open(filename, 'wb') as file:
                    file.write(img)
                return
            return self.uiAutomatorHelper.api_instance.ui_device_screenshot_get(scale=scale,
                                                                                quality=quality,
                                                                                _preload_content=False,
                                                                                **kwargs)

        def wait(self, oid: int, timeout=10000):
            """

            :param oid: the search condition ref (oid)
            :param timeout: the timeout in ms
            :return:
            """
            return self.uiAutomatorHelper.api_instance.ui_device_wait_get(oid=oid,
                                                                          timeout=timeout)

        def wait_for_idle(self, **kwargs) -> None:
            """
            Waits for idle.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs:
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_wait_for_idle_get(**kwargs))

        def wait_for_window_update(self, timeout=5000, **kwargs) -> None:
            """
            Waits for window update.

            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param timeout: the timeout
            :param kwargs:
            :return:
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_device_wait_for_window_update_get(timeout, **kwargs))

    #
    # UiObject
    #
    class UiObject(ApiBase):
        """
        Inner UiObject class.

        Notice that this class does not require an OID in its constructor.
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def exists(self, oid: int) -> bool:
            """
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid:
            :return:
            """
            response: BooleanResponse = self.uiAutomatorHelper.api_instance.ui_object_oid_exists_get(oid=oid)
            return response.value

        def get_child_count(self, oid: int) -> int:
            """
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid:
            :return:
            """
            response: NumberResponse = self.uiAutomatorHelper.api_instance.ui_object_oid_get_child_count_get(oid=oid)
            return int(response.value)

        def perform_two_pointer_gesture(self, oid: int, startPoint1: Tuple[int, int], startPoint2: Tuple[int, int],
                                        endPoint1: Tuple[int, int], endPoint2: Tuple[int, int], steps: int) -> None:
            """
            Generates a two-pointer gesture with arbitrary starting and ending points.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :param startPoint1:
            :param startPoint2:
            :param endPoint1:
            :param endPoint2:
            :param steps:
            :return: the result of the operation
            """
            body = PerformTwoPointerGestureBody(Point(startPoint1[0], startPoint1[1]),
                                                Point(startPoint2[0], startPoint2[1]),
                                                Point(endPoint1[0], endPoint1[1]),
                                                Point(endPoint2[0], endPoint2[1]),
                                                steps)
            response = self.uiAutomatorHelper.api_instance.ui_object_oid_perform_two_pointer_gesture_post(oid=oid,
                                                                                                          body=body)
            check_response(response)

        def pinch_in(self, oid: int, percentage: int, steps: int = 50) -> None:
            """
            Performs a two-pointer gesture, where each pointer moves diagonally toward the other, from the edges to the
            center of this UiObject.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :param percentage: percentage of the object's diagonal length for the pinch gesture
            :param steps: the number of steps for the gesture. Steps are injected about 5 milliseconds apart, so 100
            steps may take around 0.5 seconds to complete.
            :return: the result of the operation
            """
            response = self.uiAutomatorHelper.api_instance.ui_object_oid_pinch_in_get(oid=oid, percentage=percentage,
                                                                                      steps=steps)
            check_response(response)

        def pinch_out(self, oid: int, percentage: int, steps: int = 50) -> None:
            """
            Performs a two-pointer gesture, where each pointer moves diagonally opposite across the other, from the
            center out towards the edges of the UiObject.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :param percentage: percentage of the object's diagonal length for the pinch gesture
            :param steps: the number of steps for the gesture. Steps are injected about 5 milliseconds apart, so 100
            steps may take around 0.5 seconds to complete.
            :return: the result of the operation
            """
            response = self.uiAutomatorHelper.api_instance.ui_object_oid_pinch_out_get(oid=oid, percentage=percentage,
                                                                                       steps=steps)

            check_response(response)

    #
    # UiObject2
    #
    class UiObject2(ApiBase):
        """
        Inner UiObject2 class.

        Notice that this class does not require an OID in its constructor.
        """

        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def clear(self, oid: int) -> None:
            """
            Clears the text content if this object is an editable field.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the result of the operation
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_object2_clear_get(oid=oid))

        def click(self, oid: int) -> None:
            """
            Clicks on this object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the result of the operation
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_object2_oid_click_get(oid=oid))

        def click_and_wait(self, oid: int, event_condition_ref, timeout=10000) -> None:
            """
            Clicks and wait.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :param event_condition_ref: the event condition
            :param timeout: the timeout
            :return: the status response
            """
            check_response(
                self.uiAutomatorHelper.api_instance.ui_object2_oid_click_and_wait_get(oid, event_condition_ref,
                                                                                      timeout=timeout))

        def dump(self, oid):
            """
            Dumps the content of an object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the content
            """
            return self.uiAutomatorHelper.api_instance.ui_object2_oid_dump_get(oid=oid)

        def get_content_description(self, oid: int) -> str:
            """
            Returns the content description for this object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the text
            """
            return self.uiAutomatorHelper.api_instance.ui_object2_oid_get_content_description_get(oid=oid)

        def get_text(self, oid: int) -> str:
            """
            Returns the text value for this object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the text
            """
            return self.uiAutomatorHelper.api_instance.ui_object2_oid_get_text_get(oid=oid)

        def long_click(self, oid: int) -> None:
            """
            Performs a long click on this object.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :return: the result of the operation
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_object2_oid_long_click_get(oid=oid))

        def set_text(self, oid: int, text: str):
            """
            Sets the text content if this object is an editable field.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param oid: the oid
            :param text: the text
            :return: the result of the operation
            """
            check_response(self.uiAutomatorHelper.api_instance.ui_object2_oid_set_text_post(oid=oid, body=Text(text)))

    #
    # Until
    #
    class Until(ApiBase):
        def __init__(self, uiAutomatorHelper) -> None:
            super().__init__(uiAutomatorHelper)

        def find_object(self, **kwargs) -> ObjectRef:
            """
            Returns a SearchCondition that is satisfied when at least one element matching the selector can be found.
            The condition will return the first matching element.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs: the arguments
            :return: the search condition reference
            """
            if 'body' in kwargs:
                return self.uiAutomatorHelper.api_instance.until_find_object_post(**kwargs)
            if 'by_selector' in kwargs:
                return self.uiAutomatorHelper.api_instance.until_find_object_get(**kwargs)
            body = culebratester_client.Selector(**kwargs)
            return self.uiAutomatorHelper.api_instance.until_find_object_post(body=body)

        def find_objects(self, **kwargs) -> List[ObjectRef]:
            """
            Returns a SearchCondition that is satisfied when at least one element matching the selector can be found.
            The condition will return the first matching element.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :param kwargs: the arguments
            :return: the search condition reference
            """
            if 'body' in kwargs:
                return self.uiAutomatorHelper.api_instance.until_find_objects_post(**kwargs)
            if 'by_selector' in kwargs:
                return self.uiAutomatorHelper.api_instance.until_find_objects_get(**kwargs)
            body = culebratester_client.Selector(**kwargs)
            return self.uiAutomatorHelper.api_instance.until_find_objects_post(body=body)

        def new_window(self) -> ObjectRef:
            """
            Returns a condition that depends on a new window having appeared.
            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
            :return: the event condition reference
            """
            return self.uiAutomatorHelper.api_instance.until_new_window_get()



KEY_EVENT = dict()
KEY_EVENT['KEYCODE_SOFT_LEFT'] = 1
KEY_EVENT['KEYCODE_SOFT_RIGHT'] = 2
KEY_EVENT['KEYCODE_HOME'] = 3
KEY_EVENT['KEYCODE_BACK'] = 4
KEY_EVENT['KEYCODE_CALL'] = 5
KEY_EVENT['KEYCODE_ENDCALL'] = 6
KEY_EVENT['KEYCODE_0'] = 7
KEY_EVENT['KEYCODE_1'] = 8
KEY_EVENT['KEYCODE_2'] = 9
KEY_EVENT['KEYCODE_3'] = 10
KEY_EVENT['KEYCODE_4'] = 11
KEY_EVENT['KEYCODE_5'] = 12
KEY_EVENT['KEYCODE_6'] = 13
KEY_EVENT['KEYCODE_7'] = 14
KEY_EVENT['KEYCODE_8'] = 15
KEY_EVENT['KEYCODE_9'] = 16
KEY_EVENT['KEYCODE_STAR'] = 17
KEY_EVENT['KEYCODE_POUND'] = 18
KEY_EVENT['KEYCODE_DPAD_UP'] = 19
KEY_EVENT['KEYCODE_DPAD_DOWN'] = 20
KEY_EVENT['KEYCODE_DPAD_LEFT'] = 21
KEY_EVENT['KEYCODE_DPAD_RIGHT'] = 22
KEY_EVENT['KEYCODE_DPAD_CENTER'] = 23
KEY_EVENT['KEYCODE_VOLUME_UP'] = 24
KEY_EVENT['KEYCODE_VOLUME_DOWN'] = 25
KEY_EVENT['KEYCODE_POWER'] = 26
KEY_EVENT['KEYCODE_CAMERA'] = 27
KEY_EVENT['KEYCODE_CLEAR'] = 28
KEY_EVENT['KEYCODE_A'] = 29
KEY_EVENT['KEYCODE_B'] = 30
KEY_EVENT['KEYCODE_C'] = 31
KEY_EVENT['KEYCODE_D'] = 32
KEY_EVENT['KEYCODE_E'] = 33
KEY_EVENT['KEYCODE_F'] = 34
KEY_EVENT['KEYCODE_G'] = 35
KEY_EVENT['KEYCODE_H'] = 36
KEY_EVENT['KEYCODE_I'] = 37
KEY_EVENT['KEYCODE_J'] = 38
KEY_EVENT['KEYCODE_K'] = 39
KEY_EVENT['KEYCODE_L'] = 40
KEY_EVENT['KEYCODE_M'] = 41
KEY_EVENT['KEYCODE_N'] = 42
KEY_EVENT['KEYCODE_O'] = 43
KEY_EVENT['KEYCODE_P'] = 44
KEY_EVENT['KEYCODE_Q'] = 45
KEY_EVENT['KEYCODE_R'] = 46
KEY_EVENT['KEYCODE_S'] = 47
KEY_EVENT['KEYCODE_T'] = 48
KEY_EVENT['KEYCODE_U'] = 49
KEY_EVENT['KEYCODE_V'] = 50
KEY_EVENT['KEYCODE_W'] = 51
KEY_EVENT['KEYCODE_X'] = 52
KEY_EVENT['KEYCODE_Y'] = 53
KEY_EVENT['KEYCODE_Z'] = 54
KEY_EVENT['KEYCODE_COMMA'] = 55
KEY_EVENT['KEYCODE_PERIOD'] = 56
KEY_EVENT['KEYCODE_ALT_LEFT'] = 57
KEY_EVENT['KEYCODE_ALT_RIGHT'] = 58
KEY_EVENT['KEYCODE_SHIFT_LEFT'] = 59
KEY_EVENT['KEYCODE_SHIFT_RIGHT'] = 60
KEY_EVENT['KEYCODE_TAB'] = 61
KEY_EVENT['KEYCODE_SPACE'] = 62
KEY_EVENT['KEYCODE_SYM'] = 63
KEY_EVENT['KEYCODE_EXPLORER'] = 64
KEY_EVENT['KEYCODE_ENVELOPE'] = 65
KEY_EVENT['KEYCODE_ENTER'] = 66
KEY_EVENT['KEYCODE_DEL'] = 67
KEY_EVENT['KEYCODE_GRAVE'] = 68
KEY_EVENT['KEYCODE_MINUS'] = 69
KEY_EVENT['KEYCODE_EQUALS'] = 70
KEY_EVENT['KEYCODE_LEFT_BRACKET'] = 71
KEY_EVENT['KEYCODE_RIGHT_BRACKET'] = 72
KEY_EVENT['KEYCODE_BACKSLASH'] = 73
KEY_EVENT['KEYCODE_SEMICOLON'] = 74
KEY_EVENT['KEYCODE_APOSTROPHE'] = 75
KEY_EVENT['KEYCODE_SLASH'] = 76
KEY_EVENT['KEYCODE_AT'] = 77
KEY_EVENT['KEYCODE_NUM'] = 78
KEY_EVENT['KEYCODE_HEADSETHOOK'] = 79
KEY_EVENT['KEYCODE_FOCUS'] = 80
KEY_EVENT['KEYCODE_PLUS'] = 81
KEY_EVENT['KEYCODE_MENU'] = 82
KEY_EVENT['KEYCODE_NOTIFICATION'] = 83
KEY_EVENT['KEYCODE_SEARCH'] = 84
KEY_EVENT['KEYCODE_MEDIA_PLAY_PAUSE'] = 85
KEY_EVENT['KEYCODE_MEDIA_STOP'] = 86
KEY_EVENT['KEYCODE_MEDIA_NEXT'] = 87
KEY_EVENT['KEYCODE_MEDIA_PREVIOUS'] = 88
KEY_EVENT['KEYCODE_MEDIA_REWIND'] = 89
KEY_EVENT['KEYCODE_MEDIA_FAST_FORWARD'] = 90
KEY_EVENT['KEYCODE_MUTE'] = 91
KEY_EVENT['KEYCODE_PAGE_UP'] = 92
KEY_EVENT['KEYCODE_PAGE_DOWN'] = 93
KEY_EVENT['KEYCODE_PICTSYMBOLS'] = 94
KEY_EVENT['KEYCODE_SWITCH_CHARSET'] = 95
KEY_EVENT['KEYCODE_BUTTON_A'] = 96
KEY_EVENT['KEYCODE_BUTTON_B'] = 97
KEY_EVENT['KEYCODE_BUTTON_C'] = 98
KEY_EVENT['KEYCODE_BUTTON_X'] = 99
KEY_EVENT['KEYCODE_BUTTON_Y'] = 100
KEY_EVENT['KEYCODE_BUTTON_Z'] = 101
KEY_EVENT['KEYCODE_BUTTON_L1'] = 102
KEY_EVENT['KEYCODE_BUTTON_R1'] = 103
KEY_EVENT['KEYCODE_BUTTON_L2'] = 104
KEY_EVENT['KEYCODE_BUTTON_R2'] = 105
KEY_EVENT['KEYCODE_BUTTON_THUMBL'] = 106
KEY_EVENT['KEYCODE_BUTTON_THUMBR'] = 107
KEY_EVENT['KEYCODE_BUTTON_START'] = 108
KEY_EVENT['KEYCODE_BUTTON_SELECT'] = 109
KEY_EVENT['KEYCODE_BUTTON_MODE'] = 110
KEY_EVENT['KEYCODE_ESCAPE'] = 111
KEY_EVENT['KEYCODE_FORWARD_DEL'] = 112
KEY_EVENT['KEYCODE_CTRL_LEFT'] = 113
KEY_EVENT['KEYCODE_CTRL_RIGHT'] = 114
KEY_EVENT['KEYCODE_CAPS_LOCK'] = 115
KEY_EVENT['KEYCODE_SCROLL_LOCK'] = 116
KEY_EVENT['KEYCODE_META_LEFT'] = 117
KEY_EVENT['KEYCODE_META_RIGHT'] = 118
KEY_EVENT['KEYCODE_FUNCTION'] = 119
KEY_EVENT['KEYCODE_SYSRQ'] = 120
KEY_EVENT['KEYCODE_BREAK'] = 121
KEY_EVENT['KEYCODE_MOVE_HOME'] = 122
KEY_EVENT['KEYCODE_MOVE_END'] = 123
KEY_EVENT['KEYCODE_INSERT'] = 124
KEY_EVENT['KEYCODE_FORWARD'] = 125
KEY_EVENT['KEYCODE_MEDIA_PLAY'] = 126
KEY_EVENT['KEYCODE_MEDIA_PAUSE'] = 127
KEY_EVENT['KEYCODE_MEDIA_CLOSE'] = 128
KEY_EVENT['KEYCODE_MEDIA_EJECT'] = 129
KEY_EVENT['KEYCODE_MEDIA_RECORD'] = 130
KEY_EVENT['KEYCODE_F1'] = 131
KEY_EVENT['KEYCODE_F2'] = 132
KEY_EVENT['KEYCODE_F3'] = 133
KEY_EVENT['KEYCODE_F4'] = 134
KEY_EVENT['KEYCODE_F5'] = 135
KEY_EVENT['KEYCODE_F6'] = 136
KEY_EVENT['KEYCODE_F7'] = 137
KEY_EVENT['KEYCODE_F8'] = 138
KEY_EVENT['KEYCODE_F9'] = 139
KEY_EVENT['KEYCODE_F10'] = 140
KEY_EVENT['KEYCODE_F11'] = 141
KEY_EVENT['KEYCODE_F12'] = 142
KEY_EVENT['KEYCODE_NUM_LOCK'] = 143
KEY_EVENT['KEYCODE_NUMPAD_0'] = 144
KEY_EVENT['KEYCODE_NUMPAD_1'] = 145
KEY_EVENT['KEYCODE_NUMPAD_2'] = 146
KEY_EVENT['KEYCODE_NUMPAD_3'] = 147
KEY_EVENT['KEYCODE_NUMPAD_4'] = 148
KEY_EVENT['KEYCODE_NUMPAD_5'] = 149
KEY_EVENT['KEYCODE_NUMPAD_6'] = 150
KEY_EVENT['KEYCODE_NUMPAD_7'] = 151
KEY_EVENT['KEYCODE_NUMPAD_8'] = 152
KEY_EVENT['KEYCODE_NUMPAD_9'] = 153
KEY_EVENT['KEYCODE_NUMPAD_DIVIDE'] = 154
KEY_EVENT['KEYCODE_NUMPAD_MULTIPLY'] = 155
KEY_EVENT['KEYCODE_NUMPAD_SUBTRACT'] = 156
KEY_EVENT['KEYCODE_NUMPAD_ADD'] = 157
KEY_EVENT['KEYCODE_NUMPAD_DOT'] = 158
KEY_EVENT['KEYCODE_NUMPAD_COMMA'] = 159
KEY_EVENT['KEYCODE_NUMPAD_ENTER'] = 160
KEY_EVENT['KEYCODE_NUMPAD_EQUALS'] = 161
KEY_EVENT['KEYCODE_NUMPAD_LEFT_PAREN'] = 162
KEY_EVENT['KEYCODE_NUMPAD_RIGHT_PAREN'] = 163
KEY_EVENT['KEYCODE_VOLUME_MUTE'] = 164
KEY_EVENT['KEYCODE_INFO'] = 165
KEY_EVENT['KEYCODE_CHANNEL_UP'] = 166
KEY_EVENT['KEYCODE_CHANNEL_DOWN'] = 167
KEY_EVENT['KEYCODE_ZOOM_IN'] = 168
KEY_EVENT['KEYCODE_ZOOM_OUT'] = 169
KEY_EVENT['KEYCODE_TV'] = 170
KEY_EVENT['KEYCODE_WINDOW'] = 171
KEY_EVENT['KEYCODE_GUIDE'] = 172
KEY_EVENT['KEYCODE_DVR'] = 173
KEY_EVENT['KEYCODE_BOOKMARK'] = 174
KEY_EVENT['KEYCODE_CAPTIONS'] = 175
KEY_EVENT['KEYCODE_SETTINGS'] = 176
KEY_EVENT['KEYCODE_TV_POWER'] = 177
KEY_EVENT['KEYCODE_TV_INPUT'] = 178
KEY_EVENT['KEYCODE_STB_POWER'] = 179
KEY_EVENT['KEYCODE_STB_INPUT'] = 180
KEY_EVENT['KEYCODE_AVR_POWER'] = 181
KEY_EVENT['KEYCODE_AVR_INPUT'] = 182
KEY_EVENT['KEYCODE_PROG_RED'] = 183
KEY_EVENT['KEYCODE_PROG_GREEN'] = 184
KEY_EVENT['KEYCODE_PROG_YELLOW'] = 185
KEY_EVENT['KEYCODE_PROG_BLUE'] = 186
KEY_EVENT['KEYCODE_APP_SWITCH'] = 187
KEY_EVENT['KEYCODE_BUTTON_1'] = 188
KEY_EVENT['KEYCODE_BUTTON_2'] = 189
KEY_EVENT['KEYCODE_BUTTON_3'] = 190
KEY_EVENT['KEYCODE_BUTTON_4'] = 191
KEY_EVENT['KEYCODE_BUTTON_5'] = 192
KEY_EVENT['KEYCODE_BUTTON_6'] = 193
KEY_EVENT['KEYCODE_BUTTON_7'] = 194
KEY_EVENT['KEYCODE_BUTTON_8'] = 195
KEY_EVENT['KEYCODE_BUTTON_9'] = 196
KEY_EVENT['KEYCODE_BUTTON_10'] = 197
KEY_EVENT['KEYCODE_BUTTON_11'] = 198
KEY_EVENT['KEYCODE_BUTTON_12'] = 199
KEY_EVENT['KEYCODE_BUTTON_13'] = 200
KEY_EVENT['KEYCODE_BUTTON_14'] = 201
KEY_EVENT['KEYCODE_BUTTON_15'] = 202
KEY_EVENT['KEYCODE_BUTTON_16'] = 203
KEY_EVENT['KEYCODE_LANGUAGE_SWITCH'] = 204
KEY_EVENT['KEYCODE_MANNER_MODE'] = 205
KEY_EVENT['KEYCODE_3D_MODE'] = 206
KEY_EVENT['KEYCODE_CONTACTS'] = 207
KEY_EVENT['KEYCODE_CALENDAR'] = 208
KEY_EVENT['KEYCODE_MUSIC'] = 209
KEY_EVENT['KEYCODE_CALCULATOR'] = 210
KEY_EVENT['KEYCODE_ZENKAKU_HANKAKU'] = 211
KEY_EVENT['KEYCODE_EISU'] = 212
KEY_EVENT['KEYCODE_MUHENKAN'] = 213
KEY_EVENT['KEYCODE_HENKAN'] = 214
KEY_EVENT['KEYCODE_KATAKANA_HIRAGANA'] = 215
KEY_EVENT['KEYCODE_YEN'] = 216
KEY_EVENT['KEYCODE_RO'] = 217
KEY_EVENT['KEYCODE_KANA'] = 218
KEY_EVENT['KEYCODE_ASSIST'] = 219
KEY_EVENT['KEYCODE_BRIGHTNESS_DOWN'] = 220
KEY_EVENT['KEYCODE_BRIGHTNESS_UP'] = 221
KEY_EVENT['KEYCODE_MEDIA_AUDIO_TRACK'] = 222
KEY_EVENT['KEYCODE_SLEEP'] = 223
KEY_EVENT['KEYCODE_WAKEUP'] = 224
KEY_EVENT['KEYCODE_PAIRING'] = 225
KEY_EVENT['KEYCODE_MEDIA_TOP_MENU'] = 226
KEY_EVENT['KEYCODE_11'] = 227
KEY_EVENT['KEYCODE_12'] = 228
KEY_EVENT['KEYCODE_LAST_CHANNEL'] = 229
KEY_EVENT['KEYCODE_TV_DATA_SERVICE'] = 230
KEY_EVENT['KEYCODE_VOICE_ASSIST'] = 231
KEY_EVENT['KEYCODE_TV_RADIO_SERVICE'] = 232
KEY_EVENT['KEYCODE_TV_TELETEXT'] = 233
KEY_EVENT['KEYCODE_TV_NUMBER_ENTRY'] = 234
KEY_EVENT['KEYCODE_TV_TERRESTRIAL_ANALOG'] = 235
KEY_EVENT['KEYCODE_TV_TERRESTRIAL_DIGITAL'] = 236
KEY_EVENT['KEYCODE_TV_SATELLITE'] = 237
KEY_EVENT['KEYCODE_TV_SATELLITE_BS'] = 238
KEY_EVENT['KEYCODE_TV_SATELLITE_CS'] = 239
KEY_EVENT['KEYCODE_TV_SATELLITE_SERVICE'] = 240
KEY_EVENT['KEYCODE_TV_NETWORK'] = 241
KEY_EVENT['KEYCODE_TV_ANTENNA_CABLE'] = 242
KEY_EVENT['KEYCODE_TV_INPUT_HDMI_1'] = 243
KEY_EVENT['KEYCODE_TV_INPUT_HDMI_2'] = 244
KEY_EVENT['KEYCODE_TV_INPUT_HDMI_3'] = 245
KEY_EVENT['KEYCODE_TV_INPUT_HDMI_4'] = 246
KEY_EVENT['KEYCODE_TV_INPUT_COMPOSITE_1'] = 247
KEY_EVENT['KEYCODE_TV_INPUT_COMPOSITE_2'] = 248
KEY_EVENT['KEYCODE_TV_INPUT_COMPONENT_1'] = 249
KEY_EVENT['KEYCODE_TV_INPUT_COMPONENT_2'] = 250
KEY_EVENT['KEYCODE_TV_INPUT_VGA_1'] = 251
KEY_EVENT['KEYCODE_TV_AUDIO_DESCRIPTION'] = 252
KEY_EVENT['KEYCODE_TV_AUDIO_DESCRIPTION_MIX_UP'] = 253
KEY_EVENT['KEYCODE_TV_AUDIO_DESCRIPTION_MIX_DOWN'] = 254
KEY_EVENT['KEYCODE_TV_ZOOM_MODE'] = 255
KEY_EVENT['KEYCODE_TV_CONTENTS_MENU'] = 256
KEY_EVENT['KEYCODE_TV_MEDIA_CONTEXT_MENU'] = 257
KEY_EVENT['KEYCODE_TV_TIMER_PROGRAMMING'] = 258
KEY_EVENT['KEYCODE_HELP'] = 259
KEY_EVENT['KEYCODE_NAVIGATE_PREVIOUS'] = 260
KEY_EVENT['KEYCODE_NAVIGATE_NEXT'] = 261
KEY_EVENT['KEYCODE_NAVIGATE_IN'] = 262
KEY_EVENT['KEYCODE_NAVIGATE_OUT'] = 263
KEY_EVENT['KEYCODE_STEM_PRIMARY'] = 264
KEY_EVENT['KEYCODE_STEM_1'] = 265
KEY_EVENT['KEYCODE_STEM_2'] = 266
KEY_EVENT['KEYCODE_STEM_3'] = 267
KEY_EVENT['KEYCODE_DPAD_UP_LEFT'] = 268
KEY_EVENT['KEYCODE_DPAD_DOWN_LEFT'] = 269
KEY_EVENT['KEYCODE_DPAD_UP_RIGHT'] = 270
KEY_EVENT['KEYCODE_DPAD_DOWN_RIGHT'] = 271
KEY_EVENT['KEYCODE_MEDIA_SKIP_FORWARD'] = 272
KEY_EVENT['KEYCODE_MEDIA_SKIP_BACKWARD'] = 273
KEY_EVENT['KEYCODE_MEDIA_STEP_FORWARD'] = 274
KEY_EVENT['KEYCODE_MEDIA_STEP_BACKWARD'] = 275
KEY_EVENT['KEYCODE_SOFT_SLEEP'] = 276
KEY_EVENT['KEYCODE_CUT'] = 277
KEY_EVENT['KEYCODE_COPY'] = 278
KEY_EVENT['KEYCODE_PASTE'] = 279
KEY_EVENT['KEYCODE_SYSTEM_NAVIGATION_UP'] = 280
KEY_EVENT['KEYCODE_SYSTEM_NAVIGATION_DOWN'] = 281
KEY_EVENT['KEYCODE_SYSTEM_NAVIGATION_LEFT'] = 282
KEY_EVENT['KEYCODE_SYSTEM_NAVIGATION_RIGHT'] = 283
KEY_EVENT['KEYCODE_ALL_APPS'] = 284
KEY_EVENT['KEYCODE_REFRESH'] = 285
KEY_EVENT['KEYCODE_THUMBS_UP'] = 286
KEY_EVENT['KEYCODE_THUMBS_DOWN'] = 287
KEY_EVENT['KEYCODE_PROFILE_SWITCH'] = 288
KEY_EVENT['KEYCODE_VIDEO_APP_1'] = 289
KEY_EVENT['KEYCODE_VIDEO_APP_2'] = 290
KEY_EVENT['KEYCODE_VIDEO_APP_3'] = 291
KEY_EVENT['KEYCODE_VIDEO_APP_4'] = 292
KEY_EVENT['KEYCODE_VIDEO_APP_5'] = 293
KEY_EVENT['KEYCODE_VIDEO_APP_6'] = 294
KEY_EVENT['KEYCODE_VIDEO_APP_7'] = 295
KEY_EVENT['KEYCODE_VIDEO_APP_8'] = 296
KEY_EVENT['KEYCODE_FEATURED_APP_1'] = 297
KEY_EVENT['KEYCODE_FEATURED_APP_2'] = 298
KEY_EVENT['KEYCODE_FEATURED_APP_3'] = 299
KEY_EVENT['KEYCODE_FEATURED_APP_4'] = 300
KEY_EVENT['KEYCODE_DEMO_APP_1'] = 301
KEY_EVENT['KEYCODE_DEMO_APP_2'] = 302
KEY_EVENT['KEYCODE_DEMO_APP_3'] = 303
KEY_EVENT['KEYCODE_DEMO_APP_4'] = 304



import sys
import textwrap
from abc import ABC
from datetime import date

__version__ = '22.0.0'

from typing import TextIO, Union, Dict, List

from com.dtmilano.android.viewclient import CulebraOptions


class AbstractCodeGenerator(ABC):
    def __init__(self, options: Dict[str, Union[bool, None, str, List[str], float]]) -> None:
        self.options = options

    @staticmethod
    def get_shebang() -> str:
        return '#! /usr/bin/env python3'

    def header(self, out: TextIO = sys.stdout) -> None:
        print(textwrap.dedent(f'''
            {self.get_shebang()}
            # -*- coding: utf-8 -*-
            """
            Copyright (C) 2013-2022  Diego Torres Milano
            Created on {date.today()} by Culebra v{__version__}
                                  __    __    __    __
                                 /  \  /  \  /  \  /  \ 
            ____________________/  __\/  __\/  __\/  __\_____________________________
            ___________________/  /__/  /__/  /__/  /________________________________
                               | / \   / \   / \   / \   \___
                               |/   \_/   \_/   \_/   \    o \ 
                                                       \_____/--<
            @author: Diego Torres Milano
            @author: Jennifer E. Swofford (ascii art snake)
            """


            import os
            import re
            import sys
            import time
            ''')[1:], file=out)

    @staticmethod
    def import_unittest(out: TextIO = sys.stdout) -> None:
        print(textwrap.dedent('''
            import unittest
            '''), file=out)

    @staticmethod
    def prepend_to_syspath(out: TextIO = sys.stdout) -> None:
        print(textwrap.dedent('''
            try:
                sys.path.insert(0, os.path.join(os.environ['ANDROID_VIEW_CLIENT_HOME'], 'src'))
            except:
                pass
            '''), file=out)

    def import_viewclient(self, out: TextIO = sys.stdout) -> None:
        extra_import = ', CulebraTestCase' if self.options[CulebraOptions.UNIT_TEST_CLASS] else ''
        print(textwrap.dedent(f'''
            from com.dtmilano.android.viewclient import ViewClient, KEY_EVENT{extra_import}

            ''')[1:], file=out)

    @staticmethod
    def tag(tag: str, out: TextIO = sys.stdout) -> None:
        print(f'TAG = \'{tag}\'', file=out)

    def unittest(self, kwargs1: Dict[str, bool], kwargs2: Dict[str, Union[bool, dict]],
                 out: TextIO = sys.stdout) -> None:
        method = self.options[CulebraOptions.UNIT_TEST_METHOD] if self.options[
            CulebraOptions.UNIT_TEST_METHOD] else 'testSomething'
        print(textwrap.dedent(f'''
            class CulebraTests(CulebraTestCase):

                @classmethod
                def setUpClass(cls):
                    cls.kwargs1 = {kwargs1}
                    cls.kwargs2 = {kwargs2}
                    cls.options = {self.options}
                    cls.sleep = 5

                def setUp(self):
                    super(CulebraTests, self).setUp()

                def tearDown(self):
                    super(CulebraTests, self).tearDown()

                def preconditions(self):
                    if not super(CulebraTests, self).preconditions():
                        return False

                    if options[\'{CulebraOptions.INSTALL_APK}\']:
                        if self.vc.installPackage(options[\'{CulebraOptions.INSTALL_APK}\']) != 0:
                            return False
                            
                    return True
                    
                def {method}(self):
                    if not self.preconditions():
                        self.fail('Preconditions failed')
                        
                    _s = CulebraTests.sleep
                    _v = CulebraTests.verbose
                    pid = os.getpid()
                        
                    if options[\'{CulebraOptions.SAVE_SCREENSHOT}\']:
                        self.vc.writeImageToFile(options[\'{CulebraOptions.SAVE_SCREENSHOT}\'], 'PNG')
                        
                    if options[\'{CulebraOptions.USE_DICTIONARY}\']:
                        self.views = dict()
                    
            '''), file=out)

    # FIXME:
    # still missing these parts:
    #
    #     if not options[CulebraOptions.GUI]:
    #         vc.dump(window=options[CulebraOptions.WINDOW])
    #     indent = ' ' * 8
    #     prefix = 'self.'
    #
    #     #     if not options[CulebraOptions.DO_NOT_VERIFY_SCREEN_DUMP]:
    #     #         print '''\
    #     #         self.vc.dump(%s)
    #     #         ''' % getWindowOption()
    #     #         vc.traverse(transform=transform)
    #     #         print
    #
    #     if options[CulebraOptions.GUI]:
    #         runCulebron()
    #     elif not options[CulebraOptions.DO_NOT_VERIFY_SCREEN_DUMP]:
    #         if kwargs2.get(ViewClientOptions.USE_UIAUTOMATOR_HELPER, False):
    #             printDumpUiAutomatorHelper(getWindowOption())
    #         else:
    #             printDump(getWindowOption())
    #     else:
    #         print('''\
    #         ## your test code here ##
    #         ''')
    #
    #     print('''
    #
    # if __name__ == '__main__':
    #     CulebraTests.main()
    # ''')

    def find_object(self, view):
        raise NotImplementedError()


class HelperCodeGenerator(AbstractCodeGenerator):
    def header(self, out=sys.stdout):
        return AbstractCodeGenerator.header(self, out)

    def find_object(self, view):
        pass


class VcCodeGenerator(AbstractCodeGenerator):
    def header(self, out=sys.stdout):
        return AbstractCodeGenerator.header(self, out)

    def find_object(self, view):
        pass
